@@ -45,6 +45,10 @@ static int _is_valid_package(char* filename);
 
 static char* _access_string(char** set, int i);
 
+static int _is_info_dir(char* filename);
+static int _is_info_file(char* filename);
+static int _install_info_add(char* filename);
+static int _install_info_remove(char* filename);
 
 /* detect makefile features, and return them as an opaque type.
    Note that the structure also contains the makefile name */
@@ -113,7 +117,7 @@ SRCINST_ERR _action_init_build(struct _package_info* info, char* filename, char*
   }
   
   if (!info->name) {
-  /* set default package name, verifying uniqueness */
+    /* set default package name, verifying uniqueness */
 
     info->name = srcinst_strdup(info->packdir);
     if (_search_package_list(&_srcinst_state.packages, info->name)) {
@@ -183,14 +187,16 @@ SRCINST_ERR _action_fini_build(struct _package_info* info) {
 
   srcinst_load_cwd(_srcinst_state.dirs[SRCINST_DIR_RUNBUILD]);
   _srcinst_state.dirs[SRCINST_DIR_RUNBUILD] = 0;
-  
-  _clear_builddirs();
+
   if (info->packdir) {
     free(info->packdir); info->packdir = 0;
   }
 
   info->build_ready = 0;
 
+  if (!_clear_builddirs())
+    return SRCINST_ERR_ACTION;
+
   return SRCINST_ERR_OK;
 }
 
@@ -236,7 +242,7 @@ SRCINST_ERR _action_install_source(struct _package_info* info, SRCINST_COMP form
     return SRCINST_ERR_ACTION;
 
   info->source_location = tmp;
-  info->source_size = srcinst_file_size(tmp);
+  info->source_size = srcinst_file_size(tmp) / 1024;
   
   if (previous_source) {
     if (strcmp(info->source_location, previous_source) != 0) {
@@ -258,18 +264,11 @@ int _action_get_conflict_list(struct _package_info* info,
   if (!_get_relative_dependencies(list, info, 0))
     return 0;
   
-  for (i = 0; i < list->count; i++) {
-    int ft; size_t len;
-    ft = srcinst_file_type(list->strings[i]);
-
-    if (ft == SRCINST_TYPE_DIR ||
-
-	(len = strlen(list->strings[i])) >= 9 &&
-	strcmp(list->strings[i] + (len - 9), "/info/dir") == 0)
-
+  for (i = 0; i < list->count; i++)
+    if (srcinst_file_type(list->strings[i]) == SRCINST_TYPE_DIR ||
+	_is_info_dir(list->strings[i]))
       list->strings[i][0] = '\0';
-  }
-
+  
   srcinst_purge_string_list(list, 0);
   return 1;
 }
@@ -476,24 +475,71 @@ static int _compare_strcmp(char** str1, char** str2) {
 
 static void _strip_dirfile(struct srcinst_string_list* list) {
   int i;
-  for (i = 0; i < list->count; i++) {
-    size_t len;
-    len = strlen(list->strings[i]);
-    if (len >= 9 &&
-	strcmp(list->strings[i] + (len - 9), "/info/dir") == 0) {
-      list->strings[i][0] = 0;
-    }
-  }
 
+  for (i = 0; i < list->count; i++) {
+    if (_is_info_dir(list->strings[i]))
+      list->strings[i][0] = '\0';
+  }
+  
   srcinst_purge_string_list(list, 0);
 }
 
+static int _is_info_dir(char* filename) {
+  char* dname;
+  int rv;
+
+  dname = srcinst_dirname(filename);
+
+  rv = (strcmp(srcinst_basename(dname), "info") == 0 &&
+	strcmp(srcinst_basename(filename), "dir") == 0);
+
+  free(dname);
+  return rv;
+}
+
+static int _is_info_file(char* filename) {
+  char* dname;
+  int rv;
+
+  dname = srcinst_dirname(filename);
+  rv = strcmp(srcinst_basename(dname), "info") == 0;
+  free(dname);
+  return rv;
+}
+
+static int _install_info_add(char* filename) {
+  char* dname, *infodir;
+  int rv;
+
+  dname = srcinst_dirname(filename);
+  infodir = srcinst_strjoin("--info-dir=", dname, 0);
+
+  rv = srcinst_spawn_wait(PRG_install_info, infodir, filename, 0);
+
+  free(infodir);
+  free(dname);
+  return rv;
+}
+
+static int _install_info_remove(char* filename) {
+  char* dname, *infodir;
+  int rv = 1;
+
+  dname = srcinst_dirname(filename);
+  infodir = srcinst_strjoin("--info-dir=", dname, 0);
+    
+  rv = srcinst_spawn_wait(PRG_install_info, infodir,
+			  "--remove", filename, 0);
+  free(infodir);
+  free(dname);
+  return rv;
+}
 
 static int _internal_uninstall(struct _package_info* info) {
   struct srcinst_string_list dep; int i;
   int result = 0;
 
-  /* get a list of all file and link dependencies for the current package.
+  /* get a list of all file, dir and link dependencies for the current package.
      DEP is mirrored, not deep copied. Only (char**)->strings is alloc-ed,
      and only on success. */
   if (!_get_relative_dependencies(&dep, info, 1)) {
@@ -501,8 +547,8 @@ static int _internal_uninstall(struct _package_info* info) {
   }
 
   /* sorted, please */
-  srcinst_qsort(dep.strings, dep.count, &_compare_strcmp);
-    
+  srcinst_sort_string_list(&dep, &_compare_strcmp);
+   
   for (i = 0; i < info->installed_files.count; i++) {
     char* filename; SRCINST_TYPE ft;
     filename = info->installed_files.strings[i];
@@ -511,58 +557,42 @@ static int _internal_uninstall(struct _package_info* info) {
     if ((ft == SRCINST_TYPE_FILE || ft == SRCINST_TYPE_LINK) &&
 	srcinst_bsearch(dep.strings, dep.count, filename,
 			&_access_string) == -1) {
-
-      /* manage info files */
-      char* dname; char* bname; int remove;
-      remove = 1;
-      dname = srcinst_dirname(filename);
-      bname = srcinst_basename(dname);
-      if (strcmp(bname, "info") == 0) {
-	if (strcmp(srcinst_basename(filename), "dir") == 0) {
-	  /* _never_ _ever_ remove an info/dir file */
-	  remove = 0;
-
-	} else {
-	  char* dir;	  
-	  dir = srcinst_fnjoin(dname, "dir");
-	  if (srcinst_file_type(dir) == SRCINST_TYPE_FILE) {
-	    /* there is an info/dir file to update */
-	    char* infodir;
-	    infodir = srcinst_strjoin("--info-dir=", dname, 0);
-	    if (!srcinst_spawn_wait(PRG_install_info, 
-				    infodir, "--remove", filename, 0)) {
-	      srcinst_warning(SRCINST_ERR_ACTION,
-	     "could not cleanly remove infofile from infodir (ignoring)", 0);
-	    }
-	  }
-	  free(dir);
-	}
-      }
-
-      free(dname);
-
-      if (remove) {
-	if (!srcinst_spawn_wait(PRG_rm, "-f", filename, 0))
-	  goto end_proc;
-      }
+      int unlinkit = 1;
       
+      if (_is_info_dir(filename)) {
+	/* never _ever_ delete from the filesystem a file
+	   called info/dir */
+	unlinkit = 0;
+	
+      } else if (_is_info_file(filename)) {
+	/* calls install-info --remove, only updates the
+	   info database. Does not remove from the fs.*/
+	if (!_install_info_remove(filename))
+	  srcinst_warning(SRCINST_ERR_ACTION, 
+			  "not removed from info database", filename);
+      }
+
+      if (unlinkit && !srcinst_spawn_wait(PRG_rm, "-f", filename, 0))
+	goto end_proc;
+
       *filename = '\0'; /* schedule for removal from the string list */
     }
   }
-
+  
   srcinst_purge_string_list(&(info->installed_files), 0);
-
+  
   /* now check empty directories. First order them by path length, so we
      check contained dirs before the container using inverse `for' loop. */
 
-  srcinst_qsort(info->installed_files.strings, info->installed_files.count,
-		&_compare_dir_depth);
+  srcinst_sort_string_list(&(info->installed_files), &_compare_dir_depth);
 
   for (i = info->installed_files.count - 1; i >= 0; i--) {
     char* filename;
     filename = info->installed_files.strings[i];
 
-    if (srcinst_file_ltype(filename) == SRCINST_TYPE_DIR) {
+    if (srcinst_file_ltype(filename) == SRCINST_TYPE_DIR &&
+	srcinst_bsearch(dep.strings, dep.count, filename,
+			&_access_string) == -1) {
       struct srcinst_string_list list;
       if (!srcinst_glob_string_list(&list, filename, 0))
 	goto end_proc;
@@ -571,6 +601,7 @@ static int _internal_uninstall(struct _package_info* info) {
 	/* remove empty directory */
 	if (!srcinst_spawn_wait(PRG_rmdir, filename, 0))
 	  goto end_proc;
+
 	*filename = '\0';	/* schedule dir for removal from string list */
       }
       srcinst_free_string_list(&list);
@@ -866,8 +897,7 @@ int _detect_configure(void) {
     if (srcinst_file_access("./configure", "x"))
       return 1;
     else
-      srcinst_warning(SRCINST_ERR_PERM,
-		      "configure script is not executable", 0);
+      srcinst_warning(SRCINST_ERR_PERM, "configure script not executable", 0);
   }
   
   for (i = 0; autogen_scripts[i]; i++) {
@@ -883,7 +913,7 @@ int _detect_configure(void) {
 	    return 1;
 	  else 
 	    srcinst_warning(SRCINST_ERR_PERM,
-			    "configure script is not executable", 0);
+			    "configure script not executable", 0);
 	}
       }
     }
@@ -985,25 +1015,28 @@ static int _make(char* makefile) {
   return srcinst_spawn_wait(PRG_make, "-f", makefile, 0);
 }
 
-
-/* test installation is only possible with DESTDIR or INSTALL_ROOT support */
 static SRCINST_ERR _make_test_install(struct _package_info* info, struct _makefile_features* mf, int strip) {
 
   char* destdir_variable, *install_string;
-  int destdir_supported, install_root_supported, strip_supported;
+  int destdir_support, install_root_support, prefix_support;
+  int strip_support;
 
-  destdir_supported = mf->features[SRCINST_FEATURE_DESTDIR].supported;
-  install_root_supported = mf->features[SRCINST_FEATURE_INSTALL_ROOT].supported;
-  strip_supported = mf->features[SRCINST_FEATURE_STRIP].supported;
+  destdir_support = mf->features[SRCINST_FEATURE_DESTDIR].supported;
+  install_root_support = mf->features[SRCINST_FEATURE_INSTALL_ROOT].supported;
+  prefix_support = mf->features[SRCINST_FEATURE_PREFIX].supported;
+  strip_support = mf->features[SRCINST_FEATURE_STRIP].supported;
 
-  install_string = (strip_supported && strip ? "install-strip" : "install");
+  install_string = (strip_support && strip ? "install-strip" : "install");
 
-  if (destdir_supported) {
+  if (destdir_support) {
     destdir_variable = "DESTDIR";
 
-  } else if (install_root_supported) {
+  } else if (install_root_support) {
     destdir_variable = "INSTALL_ROOT";
 
+  } else if (prefix_support) {
+    destdir_variable = "prefix";
+
   } else {
     return SRCINST_ERR_POLICY;
   }
@@ -1036,17 +1069,24 @@ static SRCINST_ERR _make_test_install(struct _package_info* info, struct _makefi
 
     srcinst_free_string_list(&(info->installed_files));
     _strip_dirfile(&installed_files);
-    
+
+    if (strcmp(destdir_variable, "prefix") == 0) {
+      srcinst_prefix_string_list(&installed_files, info->prefix);
+    }
+
+    srcinst_sort_string_list(&installed_files, &_compare_strcmp);
+
     srcinst_mirror_string_list(&(info->installed_files), &installed_files);
     free(installed_files.strings);
 
     /* calc installation disk space usage in KB (approx) */
            
-    pattern = srcinst_spawn_wait_expect("([[:digit:]]+)", PRG_du, "-b", "-s",
+    pattern = srcinst_spawn_wait_expect("([[:digit:]]+)", PRG_du, "-k", "-s",
 					_srcinst_state.dirs[SRCINST_DIR_DEST], 0);
     
     if (pattern) {
       sscanf(pattern, OFF_FMT, &(info->installed_size));
+
     } else {
       info->installed_size = -1;
     }
@@ -1057,10 +1097,10 @@ static SRCINST_ERR _make_test_install(struct _package_info* info, struct _makefi
 
 static SRCINST_ERR _make_install(struct _package_info* info, struct _makefile_features* mf, int strip) {
   char* install_string;
-  int strip_supported;
+  int strip_support;
 
-  strip_supported = mf->features[SRCINST_FEATURE_STRIP].supported;
-  install_string = (strip_supported && strip ? "install-strip" : "install");
+  strip_support = mf->features[SRCINST_FEATURE_STRIP].supported;
+  install_string = (strip_support && strip ? "install-strip" : "install");
 
   info->installed = 1;
   info->onclose |= SRCINST_ONCLOSE_UPDATE;
@@ -1078,7 +1118,8 @@ static int _configure(char** cflags) {
   
   /* check to ensure that a makefile was generated */
   if (!(makefile = _get_makefile())) {
-    srcinst_warning(SRCINST_ERR_ACTION, "Makefile not generated by configure script", 0);
+    srcinst_warning(SRCINST_ERR_ACTION,
+		    "Makefile not generated by configure script", 0);
     return 0;
   }
 
@@ -1088,10 +1129,12 @@ static int _configure(char** cflags) {
 
 int _clear_builddirs(void) {
 
-  if (!srcinst_spawn_wait(PRG_rm, "-fr", _srcinst_state.dirs[SRCINST_DIR_BUILD], 0))
+  if (!srcinst_spawn_wait(PRG_rm, "-fr", 
+			  _srcinst_state.dirs[SRCINST_DIR_BUILD], 0))
     return 0;
 
-  if (!srcinst_spawn_wait(PRG_rm, "-fr", _srcinst_state.dirs[SRCINST_DIR_DEST], 0))
+  if (!srcinst_spawn_wait(PRG_rm, "-fr",
+			  _srcinst_state.dirs[SRCINST_DIR_DEST], 0))
     return 0;
 
   return srcinst_check_dirs();
@@ -1108,50 +1151,49 @@ SRCINST_ERR _check_build_ready(struct _package_info* info) {
 }
 
 SRCINST_ERR _action_upgrade(struct _package_info* old_info, struct _package_info* new_info) {
-
   int i;
 
+  /* remember to update when we close the old package */
+  old_info->onclose |= SRCINST_ONCLOSE_UPDATE;
+
   if (old_info->installed) {
-    if (!_internal_uninstall(old_info))
-      return SRCINST_ERR_ACTION;
+    if (!_internal_uninstall(old_info)) {
+      srcinst_warning(SRCINST_ERR_ACTION,
+		      "could not uninstall old package", 0);
+    }
     
     old_info->installed = 0;
     old_info->installed_size = (off_t)-1;
     srcinst_free_string_list(&(old_info->installed_files));
-    old_info->onclose |= SRCINST_ONCLOSE_UPDATE;
   }
   
   if (old_info->source_location) {
-    SRCINST_ERR rv;
-    if ((rv = _action_uninstall_source(old_info)) != SRCINST_ERR_OK)
-      return rv;
+
+    if (strcmp(old_info->source_location, new_info->source_location) != 0) {
+      SRCINST_ERR rv;
+      if ((rv = _action_uninstall_source(old_info)) != SRCINST_ERR_OK) {
+	srcinst_warning(SRCINST_ERR_ACTION, 
+			"could not uninstall old source", 0);
+      }
+      
+      if (old_info->source_location) {
+	/* if source uninstall fails, we _must_ force abandoning old source */
+	free(old_info->source_location);
+	old_info->source_location = 0;
+      }
+    }
   }
 
   for (i = 0; i < new_info->installed_files.count; i++) {
-    char* dname; char* bname;
+    char* filename;
+    filename = new_info->installed_files.strings[i];
 
-    dname = srcinst_dirname(new_info->installed_files.strings[i]);
-    bname = srcinst_basename(dname);
-
-    if (strcmp(bname, "info") == 0) {
-      char* dir;	  
-      dir = srcinst_fnjoin(dname, "dir");
-      if (srcinst_file_type(dir) == SRCINST_TYPE_FILE) {
-	/* there is an info/dir file to update */
-	char* infodir;
-	infodir = srcinst_strjoin("--info-dir=", dname, 0);
-	if (!srcinst_spawn_wait(PRG_install_info, infodir,
-				new_info->installed_files.strings[i], 0)) {
-	  srcinst_warning(SRCINST_ERR_ACTION,
-			  "could not upgrade info dir (ignoring)", 0);
-	}
-	free(infodir);
-      }
-      free(dir);
-    }
-    free(dname);
+    if (_is_info_file(filename))
+      if (!_install_info_add(filename))
+	srcinst_warning(SRCINST_ERR_ACTION, "could not upgrade db for info",
+			filename);
   }
-
+  
   return SRCINST_ERR_OK;
 }
 
