@@ -347,10 +347,12 @@ int srcinst_check_dirs(void) {
     t = srcinst_file_type(dirname);
 
     if (t != SRCINST_TYPE_NONE) {
+
       if (t != SRCINST_TYPE_DIR || !srcinst_file_access(dirname, "rwx")) {
 	srcinst_warning(SRCINST_ERR_CORE, "cannot access directory", dirname);
 	return 0;
       }
+
     } else if (mkdir(dirname, 0755) != 0 || !srcinst_chown_user(dirname)) {
       srcinst_warning(SRCINST_ERR_CORE, "cannot create directory", dirname);
       return 0;
@@ -483,49 +485,96 @@ SRCINST_PHANDLE srcinst_new_package(char* name) {
   return e;
 }
 
-int srcinst_close_package(SRCINST_PHANDLE h) {
-  struct _package_element* e; int rv = 1;
-  e = h;
+/* newname can be (char*)0 here too */
 
-  if (e->info->build_ready) {
-    _action_fini_build(e->info);
+SRCINST_PHANDLE srcinst_upgrade_package(char* newname, char* oldname) {
+  SRCINST_PHANDLE newh, oldh;
+  struct _package_element* olde, *newe;
+
+  if (!(oldh = srcinst_open_package(oldname)))
+    return 0;
+  
+  olde = oldh;
+
+  if (!olde->info->installed) {
+    return 0;
   }
 
-  if (e->info->onclose & SRCINST_ONCLOSE_ADD) {
-    if (e->info->installed || e->info->source_location) {
-      _add_package_element(&_srcinst_state.packages, e);
-      e->info->loaded = 1;
+  _detach_package_list(&_srcinst_state.packages, olde);
+
+  if (!(newh = srcinst_new_package(newname)))
+    goto err_proc;
+
+  newe = newh;
+  newe->info->upgrading_from = olde;
+  
+  newe->info->long_description = olde->info->long_description ?
+    strdup(olde->info->long_description) : 0;
+
+  newe->info->description = olde->info->description ?
+    strdup(olde->info->description) : 0;
+
+  return newh;
+
+ err_proc:
+  assert(_add_package_element(&_srcinst_state.packages, olde));
+  return 0;
+}
+
+int srcinst_close_package(SRCINST_PHANDLE h) {
+  struct _package_element* e;
+  struct _package_info* info;
+  int rv = 1;
+  e = h; info = e->info;
+
+  if (info->build_ready) {
+    SRCINST_ERR result;
+    if ((result = _action_fini_build(info)) != SRCINST_ERR_OK) {
+      srcinst_warning(result, "could not finalize build on package close", 0); 
+    }
+  } else {
+    if (!_clear_builddirs()) {
+      srcinst_warning(SRCINST_ERR_ACTION, "could not clear build dirs "
+		      "on package close", 0);
+    }
+  }
+
+  if (info->onclose & SRCINST_ONCLOSE_ADD) {
+
+    if ((!info->upgrading_from &&
+	 (info->installed || info->source_location)) ||
+	(info->upgrading_from && info->installed)) {
+
+      /* add package successful */
+      assert(_add_package_element(&_srcinst_state.packages, e));
+      info->loaded = 1;
       /* the action that lead to ->installed or ->source_location must
 	 have set ONCLOSE_UPDATE */
-      assert(e->info->onclose & SRCINST_ONCLOSE_UPDATE);
+      assert(info->onclose & SRCINST_ONCLOSE_UPDATE);
+      
+      if (info->upgrading_from) {
+	_action_upgrade(info->upgrading_from->info, info);
 
+	_free_package_element(info->upgrading_from);
+	free(info->upgrading_from);
+	info->upgrading_from = 0;
+      }
+      
     } else {
+      /* failed to add new package */
+      if (info->upgrading_from) {
+	assert(_add_package_element(&_srcinst_state.packages,
+				    e->info->upgrading_from));
+      }
       _free_package_element(e); free(e);
-      /* failed to add. Free memory about package and stop here. */
       return 1;
     }
   }
-  
-  if (e->info->onclose & SRCINST_ONCLOSE_UPDATE) {
+
+  if (info->onclose & SRCINST_ONCLOSE_UPDATE) {
     rv = _update_package(&_srcinst_state.packages, e);
   }
-  return rv;
-}
 
-/* interface to actions */
-
-/* remove pending stuff from a package that has been upgraded. */
-
-SRCINST_ERR srcinst_upgrade_package(SRCINST_PHANDLE oh, SRCINST_PHANDLE newh) {
-  SRCINST_ERR rv; struct _package_element* old, *new;
-  old = oh; new = newh;
-
-  if (!srcinst_close_package(newh))
-    return SRCINST_ERR_ACTION;
-
-  SRCINST_INTR_ACTION(_action_upgrade(old->info, new->info));
-
-  srcinst_close_package(old);
   return rv;
 }
 
