@@ -1,4 +1,4 @@
-/* Source Installer, Copyright (c) 2005 Claudio Fontana
+/* Source Installer, Copyright (c) 2005-2006 Claudio Fontana
 
  srcinst.h - main interface to source installer library
 
@@ -148,9 +148,9 @@ typedef enum {
 char* srcinst_get_dir(int idx);	  /* get an internal directory (see codes^)*/
 char* srcinst_get_file(int idx);  /* get an internal filename */
 
-void srcinst_log_stdout(int do_log); /* en/dis-able log of processes' output */
-void srcinst_log_stderr(int do_log); /* en/dis-able log of processes' errors */
-void srcinst_log_spawn(int do_log); /* en/dis-able log of spawned command line */
+void srcinst_log_stdout(int do_log); /* en/disable log of processes' output */
+void srcinst_log_stderr(int do_log); /* en/disable log of processes' errors */
+void srcinst_log_spawn(int do_log); /* en/disable log of spawned cmdline */
 
 /* fetch a new string list of known packages. */
 void srcinst_list_packages(struct srcinst_string_list*);
@@ -168,8 +168,17 @@ SRCINST_PHANDLE srcinst_open_package(char* name);
  */
 SRCINST_PHANDLE srcinst_new_package(char* name);
 
-/* release a package after messing with it. DO IT ALWAYS or your changes will not
-   be guaranteed to be flushed. Watch for a 0 (error) return value. */
+/* create a new package, upgrading from an old one.
+   NEWNAME can be (char*)0, to mean "use default name".
+   The function fails with 0.
+   
+   The upgrade is handled internally.
+ */
+SRCINST_PHANDLE srcinst_upgrade_package(char* newname, char* oldname);
+
+/* release a package after messing with it. 
+   DO IT ALWAYS or your changes will not be guaranteed to be flushed.
+   Watch for a 0 (error) return value. */
 int srcinst_close_package(SRCINST_PHANDLE);
 
 /* before building anything, after opening or creating the package,
@@ -247,6 +256,7 @@ typedef void* SRCINST_MFEATS;
 typedef enum {
   SRCINST_FEATURE_DESTDIR, 
   SRCINST_FEATURE_INSTALL_ROOT,
+  SRCINST_FEATURE_PREFIX,
   SRCINST_FEATURE_UNINSTALL,
   SRCINST_FEATURE_STRIP,
   SRCINST_FEATURE_N
@@ -347,16 +357,6 @@ SRCINST_ERR srcinst_export_binary(SRCINST_PHANDLE h, char* filename, SRCINST_COM
 SRCINST_ERR srcinst_rename_package(SRCINST_PHANDLE h, char* newname);
 
 
-/* the following function is to be called on an obsolete package,
-   when it has ALREADY been substituted by a newer version.
-   The function cleans up the pending files that are not in the new version,
-   and removes redundant information and possibly source code.
-   When the package is closed, it will be removed.
- */
-
-SRCINST_ERR srcinst_upgrade_package(SRCINST_PHANDLE h, SRCINST_PHANDLE to);
-
-
 /* Access preferences */
 /* always return internal strings or data; do not change pointed data or free. */
 
